/* replication of Gourinchas and Parker (Econometrica 2002) */
/* sam schulhofer-wohl, 9/13/2013 */
#ifndef GP_SSW_H
#define GP_SSW_H

/* prefix for profile output files */
#define datestr "20130918d"

/* number of periods */
#define GP_SSW_T 40

/* number of simulations */
#define GP_SSW_NSIMS 20000

/* number of threads to use */
#define GP_SSW_NUMTHREADS 8

/* number of points for gauss-hermite integration */
#define GP_SSW_NINTEGRATE 12

/* maximum iterations of the root finder for finding the consumption rule */
#define GP_SSW_RULEMAXITERS 100000

/* ruletol: tolerance on x for finding the consumption rule */
#define GP_SSW_RULETOL 1e-6

/* tolerance for numerical gradients */
#define GP_SSW_GRADTOL 1e-4

/* ngrid: number of grid points on the cash-on-hand grid */
/* ngridlow: of the grid points, number to use for tighter spacing at low end */
/* xmax: highest x on cash-on-hand grid */
/* xmin: lowest x on cash-on-hand grid */
/* xint: breakpoint between tighter and looser spacing of cash-on-hand grid points */
#define GP_SSW_NGRID 100
#define GP_SSW_NGRIDLOW 50
#define GP_SSW_XINT 2.0
#define GP_SSW_XMAX 40.0
#define GP_SSW_XMIN 1e-5

/* minimization algorithm */
#define GP_SSW_MINALGORITHM1 NLOPT_LN_BOBYQA /* first algorithm to try */
#define GP_SSW_MINALGORITHM2 NLOPT_LN_BOBYQA /* second algorithm to try */
#define GP_SSW_MINXTOL 1e-5 /* relative tolerance on x */
#define GP_SSW_MIN_MAXEVALS 10000 /* maximum function evaluations */

/* bounds for optimization */
#define GP_SSW_RHO_LB 0 /* lower bound for rho */
#define GP_SSW_RHO_UB 10 /* upper bound for rho */
#define GP_SSW_BETA_TIMES_R_UB 0.99999 /* upper bound for beta*R */
#define GP_SSW_GAMMA1_UB 1 /* upper bound for gamma1 */

/* end of settings. no changes needed below this point. */

/* helpful macros */

#if GP_SSW_NUMTHREADS>1
#define GP_SSW_PARALLEL
#else
#undef GP_SSW_PARALLEL
#endif

#define GP_SSW_NINTEGRATESQ (GP_SSW_NINTEGRATE*GP_SSW_NINTEGRATE)
#define GP_SSW_DELTALOW ((GP_SSW_XINT-GP_SSW_XMIN)/GP_SSW_NGRIDLOW)
#define GP_SSW_INVDELTALOW (1.0/GP_SSW_DELTALOW)
#define GP_SSW_DELTAHIGH ((GP_SSW_XMAX-GP_SSW_XINT)/(GP_SSW_NGRID-GP_SSW_NGRIDLOW-1))
#define GP_SSW_INVDELTAHIGH (1.0/GP_SSW_DELTAHIGH)

/* functions and structures */

int readfile(const char *filename, double *data);
int readfile(const char *filename, int *data);

struct params1 {
  double p;
  double *g;
  double R;
  double *f0;
  double s_u, s_n, wbar26, s_w26;
  double *y;
  /* R: interest rate */
  /* p: probability of zero-income shock */
  /* s_u: standard deviation of transitory log-normal income shock */
  /* s_n: standard deviation of permanent log-normal income shock */
  /* wbar26: mean of log-normal distribution of income at age 26 */
  /* s_w26: standard deviation of log-normal distribution of income at age 26 */
  /* y: income profile, 41x1 */
  /* g: growth rate of y, 40x1 */
  /* f0: smoothed profile of growth in family size, 40x1 */
  
};

struct params2 {
  double beta, rho, gamma[2], ctrend, ytrend, ftrend;
  /* beta: discount factor */
  /* rho: CRRA */
  /* gamma0, gamma1: retirement consumption function intercept and slope */
  /* ctrend: trend to add to observed consumption profile */
  /* ytrend: trend to add to observed income profile */
  /* ftrend: trend to add to observed family size profile */
};

struct problem {
  int useweights, nparams1, computeoptctrend;
  double itm, taui, gradtol_beta, gradtol_rho, gradtol_gamma0, gradtol_gamma1, gradtol_ytrend, gradtol_ftrend, gradtol_R, gradtol_p, gradtol_s_u, gradtol_s_n, optcden;
  double *intgrid, *optcnum;
  double *lcprofile;
  double *weightmat, *omegamat, *V1;
  /* nsims: number of people to simulate */
  /* nparams1: number of first-stage parameters that are treated as estimated (should = 4) */
  /* cprofile: consumption profile, Tx1 */
  /* lcprofile: log(cprofile) */
  /* optcnum:  - weightmat* (0.5*(T-1)-t) (set up using gp_ssw_optcsetup) */
  /* optcden:  (0.5*(T-1)-t)'*weightmat*(0.5*(T-1)-t) (set up using gp_ssw_optcsetup) */
  /* weightmat: optimal weighting matrix, TxT */
  /* omegamat: variance-covariance matrix of the moment conditions, TxT */
  /* V1: variance-covariance matrix of the first-stage parameters R, p, s_u, s_n, wbar26 */
  /* itm, taui: calculated while calculating the weight matrix */
  /* gradtol: tolerance for numerical gradients */
  /* useweights: =1 if use the weighting matrix in computing the objective function */
  /* computeoptctrend: =1 to compute optimal ctrend internally */
};

struct problempart2 {
  double xgridleftinv, xgridrightinv;
  double * restrict xgrid, * restrict intweights, * restrict intweightscross;
  /* xgrid: holds the cash-on-hand grid, ngrid X 1 */
  /* intweights: gauss-hermite weights, nintegrate x 1 */
  /* intweightscross: intweights * intweights' , nintegrate x nintegrate row-major */
  /* xgridleftinv=1/xgrid[0], xgridrightinv=1/xgrid[GP_SSW_NGRID-1] */

};

struct siminputs {
  double * restrict U, * restrict N, * restrict V, * restrict w26;
  /* U: log-normal transitory income shock with sd 1 (T X nsims row-major) */
  /* N: log-normal permanent income shock with sd 1 (Tx nsims row-major) */
  /* V: uniform shock to use for whether the zero-income shock occurred (T x nsims row-major) */
  /* w26: initial wealth distribution with mean 0, sd 1 (nsims X 1) */
};

struct eulerprecalc {
  /* everything here is calculated in gp_ssw_rule for use in euler_1 */
  double * restrict gXexpconst2Xgrid;
  double * restrict expconst3Xgrid;
  double ratioleft, ratioright;
  double * restrict rulediffs;
  double * restrict rulerow;
  double negrho, R_div_gf;
  double * restrict invexpconst2Xgrid;
  double I2mult, I1mult, Rbeta, g1R, em1, c1, c11, c2, c21, c3;
};

struct dist2params {
  struct problem myproblem;
  struct problempart2 myproblempart2;
  struct params1 params_stage1;
  struct siminputs simshocks;
  int optytrend, optftrend, optctrend, evalsincludinggrad;
  double copyctrend, copyytrend, copyftrend, best;
};

struct gradtol {
  double beta, rho, gamma0, gamma1, ytrend, ftrend, R, p, s_u, s_n;
};

void gp_ssw_gauher(int n, double *x, double *w);

void gp_ssw_sim(const struct problempart2 myproblempart2, const struct siminputs simshocks, const struct params1 params_stage1, const struct params2 params_stage2, double *rule, double *cm, double *ym, int calcym);

void gp_ssw_rule(const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, double *rule);

double gp_ssw_dist2(const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, const struct siminputs simshocks, double *weightmat);

double gp_ssw_du(const double c, const double negr0);

double __declspec(const) euler_1(const double cs, const double x, const struct problempart2 myproblempart2, const struct eulerprecalc precalc);

double __declspec(const) euler_last(const double cs, const double x, const double g, const struct params2 params_stage2, const struct eulerprecalc precalc); 

void gp_ssw_momentvec(const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, const struct siminputs simshocks, double *moments);

void gp_ssw_dograd(const char par, const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, const struct siminputs simshocks, const struct gradtol tols, double * gradout);

void gp_ssw_safechol(double * mat, const char * msg, const int dim);

void gp_ssw_var2(const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, const struct siminputs simshocks, const int opw, const int estctrend, const int estytrend, const int estftrend, double *grad, double *var2, double *varall, double *efficientweights, double *oid2, double *oidall);

double gp_ssw_dist2_wrapper(unsigned n, const double *x, double *grad, void *params);

void gp_ssw_nloptdecode(int minresult);

struct params2 gp_ssw_optimize(const struct problem myproblem_in, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2_in, const struct siminputs simshocks, const int estctrend, const int estytrend, const int estftrend, double *efficientweights, const int opw, double * fval, const char *profile_filename);

double gp_ssw_optctrend(const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, const struct siminputs simshocks);

void gp_ssw_optcsetup(const struct problem myproblem, double * optcnum, double * optcden);

void gp_ssw_outputprofiles(const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, const struct siminputs simshocks, const char *filename);

void gp_ssw_outputweightmat(const struct problem myproblem, const char *filename);

#endif
