/* replication of Gourinchas and Parker (Econometrica 2002) */
/* sam schulhofer-wohl, 9/13/2013 */
/* minimizes the objective function */
#include <iostream>
#include <math.h>
#include <nlopt.h>
#include <omp.h>
#include <mkl.h>
#include "gp_ssw_intel.h"
using namespace std;

void gp_ssw_nloptdecode(int minresult) {
  if(minresult<0) {
    printf("nlopt failed with return code %d ",minresult);
    if(minresult==NLOPT_FAILURE) printf("(generic failure code).\n");
    else if(minresult==NLOPT_INVALID_ARGS) printf("(invalid arguments).\n");
    else if(minresult==NLOPT_OUT_OF_MEMORY) printf("(out of memory).\n");
    else if(minresult==NLOPT_ROUNDOFF_LIMITED) printf("(halted because roundoff errors limited progress).\n");
    else if(minresult==NLOPT_FORCED_STOP) printf("(forced termination by user).\n");
    else printf("(unknown return code).\n");
  }
  else {
    printf("nlopt succeeded with return code %d ",minresult);
    if(minresult==NLOPT_SUCCESS) printf("(generic success return value).\n");
    else if(minresult==NLOPT_XTOL_REACHED) printf("(optimization stopped because xtol_rel or xtol_abs was reached).\n");
    else if(minresult==NLOPT_MAXEVAL_REACHED) printf("(optimization stopped because maxeval was reached).\n");
    else printf("(unknown return code).\n");
  }
}

/* params_stage2_in should contain the starting point */
/* on exit, it contains the optimal parameters */
/* efficientweights is TxT. must exist on entry. on exit, contains the efficient weighting matrix implied by the current estimates. */
/* opw:=1 if myproblem.weightmat has been set to the optimal weighting matrix, =0 otherwise. (passed to gp_ssw_var2). */
/* fval: value of objective function at optimum will be stored here on exit */
/* profile_filename: place to write profiles */

struct params2 gp_ssw_optimize(const struct problem myproblem_in, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2_in, const struct siminputs simshocks, const int estctrend, const int estytrend, const int estftrend, double *efficientweights, const int opw, double * fval, const char *profile_filename)
{

  struct problem myproblem __attribute__((aligned));
  myproblem=myproblem_in;
  int i,j,nparams_inner,nparams_outer;
  double * optcnum;
  optcnum=(double *)mkl_malloc(GP_SSW_T*sizeof(double),64);
  double optcden;
  struct dist2params dist2_wrapper_params __attribute__((aligned));
  dist2_wrapper_params.evalsincludinggrad=0;
  nparams_inner=4;
  nparams_outer=4;
  if(estctrend==1) {
    gp_ssw_optcsetup(myproblem,optcnum,&optcden);
    myproblem.computeoptctrend=1;
    myproblem.optcnum=optcnum;
    myproblem.optcden=optcden;
    dist2_wrapper_params.optctrend=1;
    nparams_outer++;
  }
  else {
    myproblem.computeoptctrend=0;
    dist2_wrapper_params.optctrend=0;
    dist2_wrapper_params.copyctrend=params_stage2_in.ctrend;
    }
  if(estytrend==1) {
    dist2_wrapper_params.optytrend=1;
    nparams_outer++;
    nparams_inner++;
  }
  else {
    dist2_wrapper_params.optytrend=0;
    dist2_wrapper_params.copyytrend=params_stage2_in.ytrend;
  }
  if(estftrend==1) {
    dist2_wrapper_params.optftrend=1;
    nparams_outer++;
    nparams_inner++;
  }
  else {
    dist2_wrapper_params.optftrend=0;
    dist2_wrapper_params.copyftrend=params_stage2_in.ftrend;
  }
  dist2_wrapper_params.myproblem=myproblem;
  dist2_wrapper_params.myproblempart2=myproblempart2;
  dist2_wrapper_params.params_stage1=params_stage1;
  dist2_wrapper_params.simshocks=simshocks;

  /* starting point */
  double * params_stage2_vector;
  params_stage2_vector=(double *)mkl_malloc(nparams_inner*sizeof(double),64);
  params_stage2_vector[0]=params_stage2_in.beta;
  params_stage2_vector[1]=params_stage2_in.rho;
  params_stage2_vector[2]=params_stage2_in.gamma[0]/params_stage2_in.gamma[1]; /* we optimize over gamma0/gamma1 and gamma1 so we can easily impose gamma0<gamma1 */
  params_stage2_vector[3]=params_stage2_in.gamma[1];
  int nextparam=4;
  if(estytrend==1) {
    params_stage2_vector[nextparam]=params_stage2_in.ytrend;
    nextparam++;
  }
  if(estftrend==1) {
    params_stage2_vector[nextparam]=params_stage2_in.ftrend;
    }
  dist2_wrapper_params.best=1e6;
  printf("nparams_inner: %d\n",nparams_inner);
  printf("dist2_wrapper_params.optytrend: %d\n",dist2_wrapper_params.optytrend);
  printf("starting point:\n");
  gp_ssw_dist2_wrapper(1,params_stage2_vector,NULL,&dist2_wrapper_params);

  /* define bounds */  
  double lb[nparams_inner], ub[nparams_inner];
  lb[0]=0; ub[0]=/* GP_SSW_BETA_TIMES_R_UB/params_stage1.R */ 1; /* beta */
  lb[1]=GP_SSW_RHO_LB; ub[1]=GP_SSW_RHO_UB; /* rho */
  lb[2]=0; ub[2]=1; /* gamma0/gamma1 */
  lb[3]=0; ub[3]=GP_SSW_GAMMA1_UB; /* gamma1 */
  if(nparams_inner>4) {
    for(j=4;j<nparams_inner;j++) {
      lb[j]=-1; ub[j]=1;
      }
    }

  /* minimize */
  nlopt_opt opt;
  int minresult;
  double test_timer_start = omp_get_wtime();
  /* four attempts, two alternating algorithms */
  for(j=1;j<=2;j++) {
    if(j==1) opt=nlopt_create(GP_SSW_MINALGORITHM1,nparams_inner);
    else if(j==2) opt=nlopt_create(GP_SSW_MINALGORITHM2,nparams_inner);
    nlopt_set_lower_bounds(opt,lb);
    nlopt_set_upper_bounds(opt,ub);
    nlopt_set_min_objective(opt,gp_ssw_dist2_wrapper,&dist2_wrapper_params);
    nlopt_set_xtol_rel(opt,GP_SSW_MINXTOL);
    nlopt_set_maxeval(opt,GP_SSW_MIN_MAXEVALS); 
    minresult=nlopt_optimize(opt,params_stage2_vector,fval);
    printf("attempt %d: ",j);
    gp_ssw_nloptdecode(minresult);
    cout << flush;
    nlopt_destroy(opt);
    }
  printf("ending point:\n");
  gp_ssw_dist2_wrapper(1,params_stage2_vector,NULL,&dist2_wrapper_params);
  double test_timer_end = omp_get_wtime();
  printf("time to optimize: %g seconds\n",test_timer_end-test_timer_start);  
  
  /* print results and save the minimizer */
  struct params2 params_stage2_out __attribute__((aligned));
  printf("parameters at result:\n");
  params_stage2_out.beta=params_stage2_vector[0];
  printf("beta   = %# 15.14g\n",params_stage2_out.beta);
  params_stage2_out.rho=params_stage2_vector[1];
  printf("rho    = %# 15.14g\n",params_stage2_out.rho);
  params_stage2_out.gamma[0]=params_stage2_vector[2]*params_stage2_vector[3];
  printf("gamma0 = %# 15.14g\n",params_stage2_out.gamma[0]);
  params_stage2_out.gamma[1]=params_stage2_vector[3];
  printf("gamma1 = %# 15.14g\n",params_stage2_out.gamma[1]);
  if(estctrend==1) {
    params_stage2_out.ctrend=gp_ssw_optctrend(myproblem,myproblempart2,params_stage1,params_stage2_out,simshocks);
    printf("ctrend = %# 15.14g\n",params_stage2_out.ctrend);
  }
  else {
    params_stage2_out.ctrend=params_stage2_in.ctrend;
  }
  nextparam=4;
  if(estytrend==1) {
    params_stage2_out.ytrend=params_stage2_vector[nextparam];
    printf("ytrend = %# 15.14g\n",params_stage2_out.ytrend);
    nextparam++;
  }
  else {
    params_stage2_out.ytrend=params_stage2_in.ytrend;
  }
  if(estftrend==1) {
    params_stage2_out.ftrend=params_stage2_vector[nextparam];
    printf("ftrend = %# 15.14g\n",params_stage2_out.ftrend);
  }
  else {
    params_stage2_out.ftrend=params_stage2_in.ftrend;
  }
  /* turn off computeoptctrend for the variance calculation */
  myproblem.computeoptctrend=0;
  double * grad, * var2, * varall;
  grad=(double *)mkl_malloc(GP_SSW_T*nparams_outer*sizeof(double),64);
  var2=(double *)mkl_malloc(nparams_outer*nparams_outer*sizeof(double),64);
  varall=(double *)mkl_malloc(nparams_outer*nparams_outer*sizeof(double),64);
  double oid2, oidall;
  cout << flush;
  gp_ssw_var2(myproblem,myproblempart2,params_stage1,params_stage2_out,simshocks,opw,estctrend,estytrend,estftrend,grad,var2,varall,efficientweights,&oid2,&oidall);
  printf("variance-covariance matrix at optimum (treating first stage as fixed):\n");
  for(i=0;i<nparams_outer;i++) {
    for(j=0;j<nparams_outer;j++) printf("%# 15.14g\t",var2[i*nparams_outer+j]);
    printf("\n");
  }
  printf("chi-squared at optimum (treating first stage as fixed) = %# 15.14g\n",oid2);
  printf("variance-covariance matrix at optimum (treating first stage as estimated):\n");
  for(i=0;i<nparams_outer;i++) {
    for(j=0;j<nparams_outer;j++) printf("%# 15.14g\t",varall[i*nparams_outer+j]);
    printf("\n");
  }
  printf("standard errors (treating first stage as estimated):\n");
  for(i=0;i<nparams_outer;i++) {
    printf("%# 15.14g\n",sqrt(varall[i*nparams_outer+i]));
    }
  printf("chi-squared at optimum (treating first stage as estimated) = %# 15.14g\n",oidall);
  cout << flush;
  
  /* output profiles */
  gp_ssw_outputprofiles(myproblem,myproblempart2,params_stage1,params_stage2_out,simshocks,profile_filename);
  
  
  mkl_free(grad);
  mkl_free(var2);
  mkl_free(varall);
  mkl_free(params_stage2_vector);
  mkl_free(optcnum);
  
  return params_stage2_out;
  
  
}

