/* replication of Gourinchas and Parker (Econometrica 2002) */
/* sam schulhofer-wohl, 9/13/2013 */
/* computes the consumption rules */
#include <mkl.h>
#include <math.h>
#include <float.h>
#include "gp_ssw_intel.h"
#define M_1_SQRTPI 0.56418958354776

/* rule: (T+1)xngrid. must exist on input. on output, filled with consumption rules */

void gp_ssw_rule(const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, double *rule)
{

  int i, j;
  double g;
  
  /* calculations used repeatedly in the euler equations */
  double expconst2Xgrid[GP_SSW_NINTEGRATE] __attribute__((aligned(16)));
  double invexpconst2Xgrid[GP_SSW_NINTEGRATE] __attribute__((aligned(16)));
  double expconst3Xgrid[GP_SSW_NINTEGRATE] __attribute__((aligned(16)));
  struct eulerprecalc precalc __attribute__((aligned));
  const double const2=M_SQRT2*params_stage1.s_n;
  const double const3=M_SQRT2*params_stage1.s_u;
  #pragma ivdep
  for(i=0;i<GP_SSW_NINTEGRATE;i++) {
    expconst2Xgrid[i]=exp(const2*myproblem.intgrid[i]);
  }
  vdInv(GP_SSW_NINTEGRATE,expconst2Xgrid,invexpconst2Xgrid);
  #pragma ivdep
  for(i=0;i<GP_SSW_NINTEGRATE;i++) {
    expconst3Xgrid[i]=exp(const3*myproblem.intgrid[i]);
  }
  precalc.invexpconst2Xgrid=&invexpconst2Xgrid[0];
  precalc.expconst3Xgrid=&expconst3Xgrid[0];
  double rulediffs[GP_SSW_NGRID-1];
  precalc.rulediffs=&rulediffs[0];
  double gXexpconst2Xgrid[GP_SSW_NINTEGRATE] __attribute__((aligned(16)));
  precalc.gXexpconst2Xgrid=&gXexpconst2Xgrid[0];
  precalc.Rbeta=params_stage2.beta*params_stage1.R;
  precalc.g1R=params_stage2.gamma[1]*params_stage1.R;
  precalc.em1=exp(-460.51701859881/params_stage2.rho);
  precalc.negrho=-params_stage2.rho;
  precalc.c11=pow(precalc.em1,precalc.negrho);
  precalc.c21=params_stage2.rho*precalc.c11/precalc.em1;
  precalc.c1=pow(1e-4,precalc.negrho);
  precalc.c2=params_stage2.rho*precalc.c1*1e4;
  precalc.c3=precalc.c1+(1e-4)*precalc.c2;
  precalc.I2mult=M_1_PI*(1-params_stage1.p)*precalc.Rbeta;
  precalc.I1mult=M_1_SQRTPI*params_stage1.p*precalc.Rbeta;
  
  /* final consumption function */
  int start=GP_SSW_T*GP_SSW_NGRID;
  #pragma ivdep
  for(i=0;i<GP_SSW_NGRID;i++) rule[start+i]=params_stage2.gamma[0]+params_stage2.gamma[1]*myproblempart2.xgrid[i];

  /* work backwards, solving the euler equation */
 
  /* last age */
  g=(1.0/(params_stage1.f0[GP_SSW_T-1]+params_stage2.ftrend))+params_stage2.ytrend;
  int rulerowstart=(GP_SSW_T-1)*GP_SSW_NGRID;
  /* solve at each grid point */
#ifdef GP_SSW_PARALLEL
#pragma omp parallel for default(none) private(i) shared(rule,precalc,g) num_threads(GP_SSW_NUMTHREADS) schedule(guided,1) 
#endif
  for(i=0;i<GP_SSW_NGRID;i++) {
    double newrule, xi=myproblempart2.xgrid[i];
#define FOPTS xi,g,params_stage2,precalc
    if(euler_last(1,FOPTS)>0) newrule=xi;
    /* check if credit constrained */
    else {
      const double scale=0.8;
      double x_hi, x_lo, f_hi, f_lo=1 /* irrelevant initialization to prevent compiler warning*/;
      /* if not constrained, solve the euler equation */
      /* turn initial guess into a range bracketing the zero */
      x_hi=0.5;
      f_hi=euler_last(x_hi,FOPTS);
      if(f_hi<0) {
        while(f_hi<0 && x_hi>2e-9) {
          f_lo=f_hi;
          x_hi=x_hi*scale;
          f_hi=euler_last(x_hi,FOPTS);
        }
        x_lo=x_hi/scale;
      }
      else {
        x_lo=0.5;
        f_lo=euler_last(x_lo,FOPTS);
        double onemscale=1-scale;
        while(f_lo>0) {
          f_hi=f_lo;
          x_lo=onemscale*x_lo+scale;
          f_lo=euler_last(x_lo,FOPTS);
        }
        x_hi=(x_lo-scale)/onemscale;
      }
      /* check whether we have now bracketed zero */
      if(f_lo>0) newrule=xi;
      else if(f_hi<0) newrule=(2e-9)*xi;/* very low consumption */
      else if(f_lo==0) newrule=x_lo*xi;
      else if(f_hi==0) newrule=x_hi*xi;
      else {
        /* we have bracketed zero, use brent's method (algorithm from wikipedia) to find the rule */
        int solv_iter=0;
        double brent_a, brent_b, brent_c, brent_fa, brent_fb, brent_fc, brent_s, brent_d, brent_fs, brent_temp;
        int brent_mflag=1;
        if(fabs(f_lo)<fabs(f_hi)) {
          brent_a=x_hi;
          brent_fa=f_hi;
          brent_b=x_lo;
          brent_fb=f_lo;
        }
        else {
          brent_a=x_lo;
          brent_fa=f_lo;
          brent_b=x_hi;
          brent_fb=f_hi;
        }
        brent_c=brent_a;
        brent_fc=brent_fa;
        do {
          solv_iter++;
          double brent_famfb=brent_fa-brent_fb, brent_famfc=brent_fa-brent_fc, brent_fbmfc=brent_fb-brent_fc;
          if(brent_famfc!=0 && brent_fbmfc!=0)
            brent_s=(brent_a*brent_fb*brent_fc/(brent_famfb*brent_famfc))
                   -(brent_b*brent_fa*brent_fc/(brent_famfb*brent_fbmfc))
                   +(brent_c*brent_fa*brent_fb/(brent_famfc*brent_fbmfc));
          else brent_s=brent_b+brent_fb*(brent_b-brent_a)/brent_famfb;
          double brent_check=(3*brent_a+brent_b)*0.25;
          if(
              (brent_s<brent_check && brent_s>brent_b) ||
              (brent_s>brent_check && brent_s<brent_b)
             ) {
            brent_s=0.5*(brent_a+brent_b);
            brent_mflag=1;              
          }
          else {
            brent_check=2*fabs(brent_s-brent_b);
            if(brent_mflag==1) {
              double absbmc=fabs(brent_b-brent_c);
              if(brent_check>=absbmc || absbmc<GP_SSW_RULETOL) {
                brent_s=0.5*(brent_a+brent_b);
                brent_mflag=1;
              }
              else brent_mflag=0;
            }
            else {
              double abscmd=fabs(brent_c-brent_d);
              if(brent_check>=abscmd || abscmd<GP_SSW_RULETOL) {
                brent_s=0.5*(brent_a+brent_b);
                brent_mflag=1;
              }
              else brent_mflag=0;                
            }
          }
          brent_fs=euler_last(brent_s,FOPTS);
          brent_d=brent_c;
          brent_c=brent_b;
          brent_fc=brent_fb;
          if((brent_fa>0 && brent_fs<0) || (brent_fa<0 && brent_fs>0)) {
            brent_b=brent_s;
            brent_fb=brent_fs;
          }
          else {
            brent_a=brent_s;
            brent_fa=brent_fs;
          }
          if(fabs(brent_fa)<fabs(brent_fb)) {
            brent_temp=brent_a;
            brent_a=brent_b;
            brent_b=brent_temp;
            brent_temp=brent_fa;
            brent_fa=brent_fb;
            brent_fb=brent_temp;
          }
        } while(fabs(brent_b-brent_a)>GP_SSW_RULETOL && solv_iter<GP_SSW_RULEMAXITERS);
        newrule=brent_b*xi;
      }
    }
    /* write the results to the rule matrix */
    rule[(GP_SSW_T-1) * GP_SSW_NGRID + i]=newrule;

  } /* end of loop over grid points */

  /* remaining ages */
  for(j=1;j<GP_SSW_T;j++) {
    /* update some calculations */
    g=params_stage1.g[GP_SSW_T-j-1]+params_stage2.ytrend;
    precalc.ratioleft=rule[(GP_SSW_T-j) * GP_SSW_NGRID + 0]*myproblempart2.xgridleftinv;
    precalc.ratioright=rule[(GP_SSW_T-j) * GP_SSW_NGRID + GP_SSW_NGRID-1]*myproblempart2.xgridrightinv;
    /* rulediffs = rule(1...NGRID-1) - rule(0...NGRID-2) */
    vdSub(GP_SSW_NGRID-1,&(rule[rulerowstart+1]),&(rule[rulerowstart]),precalc.rulediffs);
    #pragma ivdep
    for(i=0;i<GP_SSW_NINTEGRATE;i++) precalc.gXexpconst2Xgrid[i]=g*expconst2Xgrid[i];
    precalc.R_div_gf=params_stage1.R/(g*(params_stage1.f0[GP_SSW_T-j-1]+params_stage2.ftrend));
    precalc.rulerow=&(rule[rulerowstart]);
    rulerowstart=rulerowstart-GP_SSW_NGRID;
    /* solve at each grid point */
#ifdef GP_SSW_PARALLEL
#pragma omp parallel for default(none) private(i) shared(j,rule,precalc,g) num_threads(GP_SSW_NUMTHREADS) schedule(guided,1) 
#endif
    for(i=0;i<GP_SSW_NGRID;i++) {
      double newrule, xi=myproblempart2.xgrid[i];
#undef FOPTS
#define FOPTS xi,myproblempart2,precalc
      if(euler_1(1,FOPTS)>0) newrule=xi;
      /* check if credit constrained */
      else {
        const double scale=0.8;
        double guess, x_hi, x_lo, f_hi, f_lo=1 /* irrelevant initialization to prevent compiler warning*/;
        /* if not constrained, solve the euler equation */
        /* initial guess */
        if(fabs(rule[(GP_SSW_T-j) * GP_SSW_NGRID + i]-xi)>1e-3) guess=rule[(GP_SSW_T-j) * GP_SSW_NGRID + i]/xi;
        else guess=0.5;
        /* turn initial guess into a range bracketing the zero */
        x_hi=guess;
        f_hi=euler_1(x_hi,FOPTS);
        if(f_hi<0) {
          while(f_hi<0 && x_hi>2e-9) {
            f_lo=f_hi;
            x_hi=x_hi*scale;
            f_hi=euler_1(x_hi,FOPTS);
          }
          x_lo=x_hi/scale;
        }
        else {
          x_lo=guess;
          f_lo=euler_1(x_lo,FOPTS);
          const double onemscale=1-scale;
          while(f_lo>0) {
            f_hi=f_lo;
            x_lo=onemscale*x_lo+scale;
            f_lo=euler_1(x_lo,FOPTS);
          }
          x_hi=(x_lo-scale)/onemscale;
        }
        /* check whether we have now bracketed zero */
        if(f_lo>0) newrule=xi;
        else if(f_hi<0) newrule=(2e-9)*xi;/* very low consumption */
        else if(f_lo==0) newrule=x_lo*xi;
        else if(f_hi==0) newrule=x_hi*xi;
        else {
          /* we have bracketed zero, use brent's method (algorithm from wikipedia) to find the rule */
          int solv_iter=0;
          double brent_a, brent_b, brent_c, brent_fa, brent_fb, brent_fc, brent_s, brent_d, brent_fs, brent_temp;
          int brent_mflag=1;
          if(fabs(f_lo)<fabs(f_hi)) {
            brent_a=x_hi;
            brent_fa=f_hi;
            brent_b=x_lo;
            brent_fb=f_lo;
          }
          else {
            brent_a=x_lo;
            brent_fa=f_lo;
            brent_b=x_hi;
            brent_fb=f_hi;
          }
          brent_c=brent_a;
          brent_fc=brent_fa;
          do {
            solv_iter++;
            double brent_famfb=brent_fa-brent_fb, brent_famfc=brent_fa-brent_fc, brent_fbmfc=brent_fb-brent_fc;
            if(brent_famfc!=0 && brent_fbmfc!=0)
              brent_s=(brent_a*brent_fb*brent_fc/(brent_famfb*brent_famfc))
                     -(brent_b*brent_fa*brent_fc/(brent_famfb*brent_fbmfc))
                     +(brent_c*brent_fa*brent_fb/(brent_famfc*brent_fbmfc));
            else brent_s=brent_b+brent_fb*(brent_b-brent_a)/brent_famfb;
            double brent_check=(3*brent_a+brent_b)*0.25;
            if(
                (brent_s<brent_check && brent_s>brent_b) ||
                (brent_s>brent_check && brent_s<brent_b)
               ) {
              brent_s=0.5*(brent_a+brent_b);
              brent_mflag=1;              
            }
            else {
              brent_check=2*fabs(brent_s-brent_b);
              if(brent_mflag==1) {
                double absbmc=fabs(brent_b-brent_c);
                if(brent_check>=absbmc || absbmc<GP_SSW_RULETOL) {
                  brent_s=0.5*(brent_a+brent_b);
                  brent_mflag=1;
                }
                else brent_mflag=0;
              }
              else {
                double abscmd=fabs(brent_c-brent_d);
                if(brent_check>=abscmd || abscmd<GP_SSW_RULETOL) {
                  brent_s=0.5*(brent_a+brent_b);
                  brent_mflag=1;
                }
                else brent_mflag=0;                
              }
            }
            brent_fs=euler_1(brent_s,FOPTS);
            brent_d=brent_c;
            brent_c=brent_b;
            brent_fc=brent_fb;
            if((brent_fa>0 && brent_fs<0) || (brent_fa<0 && brent_fs>0)) {
              brent_b=brent_s;
              brent_fb=brent_fs;
            }
            else {
              brent_a=brent_s;
              brent_fa=brent_fs;
            }
            if(fabs(brent_fa)<fabs(brent_fb)) {
              brent_temp=brent_a;
              brent_a=brent_b;
              brent_b=brent_temp;
              brent_temp=brent_fa;
              brent_fa=brent_fb;
              brent_fb=brent_temp;
            }
          } while(fabs(brent_b-brent_a)>GP_SSW_RULETOL && solv_iter<GP_SSW_RULEMAXITERS);
          newrule=brent_b*xi;
        }
      }
      /* write the results to the rule matrix */
      rule[(GP_SSW_T-j-1)*GP_SSW_NGRID+i]=newrule;
    } /* end of loop over grid points */
  } /* end of loop over ages */
    
}


