/* replication of Gourinchas and Parker (Econometrica 2002) */
/* sam schulhofer-wohl, 9/13/2013 */
/* simulate consumption given consumption rule, simulated income shocks, simulated initial wealth distribution, and first-stage parameters */
#include <math.h>
#include <mkl.h>
#include <ipp.h>
#include "gp_ssw_intel.h"

/* rule: (T+1)xngrid matrix of consumption rules */
/* cm: Tx1 vector. must exist on input. on output, filled with average consumption by age. */

void gp_ssw_sim(const struct problempart2 myproblempart2, const struct siminputs simshocks, const struct params1 params_stage1, const struct params2 params_stage2, double *rule, double *cm, double *ym, int calcym)
{
  
  int i, j, age, pos, rulerowstart;
  const int datasize=GP_SSW_T*GP_SSW_NSIMS;
  double n0i, cmi, posd, ratioleft, ratioright, scale;
  
  /* create workspace */
  double * work_N, * work_NT;
  work_N=(double *)mkl_calloc(GP_SSW_NSIMS,sizeof(double),64);
  work_NT=(double *)mkl_calloc(GP_SSW_NSIMS*GP_SSW_T,sizeof(double),64);

  /* simulate initial wealth */
  double * _w_work;
  _w_work=(double *)mkl_malloc(GP_SSW_NSIMS*sizeof(double),64);
  cblas_daxpy(GP_SSW_NSIMS,params_stage1.s_w26,simshocks.w26,1,work_N,1);
  vdExp(GP_SSW_NSIMS,work_N,_w_work);
  
  /* simulate transitory shocks */
  double * _u_work;
  _u_work=(double *)mkl_malloc(GP_SSW_T*GP_SSW_NSIMS*sizeof(double),64);
  cblas_daxpy(datasize,params_stage1.s_u,simshocks.U,1,work_NT,1);
  vdExp(datasize,work_NT,_u_work);
#pragma ivdep
  for(i=0;i<datasize;i++) {
    if(simshocks.V[i]<=params_stage1.p) _u_work[i]=0;
  }

  /* simulate permanent shock */
  double * _n_work;
  _n_work=(double *)mkl_malloc(GP_SSW_T*GP_SSW_NSIMS*sizeof(double),64);
  ippsMulC_64f(simshocks.N,-params_stage1.s_n,work_NT,datasize);
  vdExp(datasize,work_NT,_n_work);
 
  /* normalized cash on hand */
  double * xsn;
  xsn=(double *)mkl_calloc(GP_SSW_T*GP_SSW_NSIMS,sizeof(double),64);
  double * xsnrow=&xsn[0], *xsnnextrow;
  double * _u_workrow=&_u_work[0];
  /* normalized cash on hand at age 0 is initial wealth + u, because initial wealth is already normalized by permanent income */
  cblas_dcopy(GP_SSW_NSIMS,_u_workrow,1,xsnrow,1);
  cblas_daxpy(GP_SSW_NSIMS,exp(params_stage1.wbar26),_w_work,1,xsnrow,1);
  mkl_free(_w_work); /* compiler does bad things if this is freed later -- don't know why */
  
  /* compute intercept so that adding a slope ytrend doesn't change mean of income profile */
  /* we add ytrend*i to element i, for i=0 to T-1 */
  /* so in total, we add ytrend*(T-1)*T/2 */
  /* that raises the mean by ytrend*(T-1)/2 */
  double yoffset=0.5*params_stage2.ytrend*((double)(GP_SSW_T-1));

  /* income profile times permanent shock */
  double * ypn;
  ypn=(double *)mkl_calloc(GP_SSW_T*GP_SSW_NSIMS,sizeof(double),64);
  double * ypnrow=&ypn[0];
  double * _n_workrow=&_n_work[0];
  /* yoffset is in logs, so divide by exp(yoffset) */
  double y0=params_stage1.y[0]/exp(yoffset);
  cblas_daxpy(GP_SSW_NSIMS,y0,_n_workrow,1,ypnrow,1);
  
  /* simulate consumption for all ages except the last */
  double * csn, * csnrow;
  csn=(double *)mkl_malloc(GP_SSW_T*GP_SSW_NSIMS*sizeof(double),64);
  double diffs[GP_SSW_NGRID-1];
  rulerowstart=0;
  for(age=0;age<GP_SSW_T-1;age++) {
    /* interpolate consumption at each value in row age of the normalized cash on hand matrix using column age of the rule matrix */
    /* compute propensity to consume at the endpoints */
    ratioleft=rule[rulerowstart+0]*myproblempart2.xgridleftinv;
    ratioright=rule[rulerowstart+GP_SSW_NGRID-1]*myproblempart2.xgridrightinv;
    /* compute differences of the rule */
    vdSub(GP_SSW_NGRID,&rule[rulerowstart+1],&rule[rulerowstart],diffs);
    /* interpolate/extrapolate */
#pragma ivdep
    for(j=0;j<GP_SSW_NSIMS;j++) {
      double x=xsn[age*GP_SSW_NSIMS+j];
      if(x<GP_SSW_XMIN) csn[age*GP_SSW_NSIMS+j]=x*ratioleft;
      else if(x>GP_SSW_XMAX) csn[age*GP_SSW_NSIMS+j]=x*ratioright;
      else if(x<GP_SSW_XINT) {
        posd=(x-GP_SSW_XMIN)*GP_SSW_INVDELTALOW;
        pos=(int)posd;
        csn[age*GP_SSW_NSIMS+j]=rule[rulerowstart+pos]+(posd-(double)pos)*diffs[pos];
      }
      else {
        posd=(x-GP_SSW_XINT)*GP_SSW_INVDELTAHIGH;
        pos=(int)posd+GP_SSW_NGRIDLOW;
        csn[age*GP_SSW_NSIMS+j]=rule[rulerowstart+pos]+(posd-(double)(pos-GP_SSW_NGRIDLOW))*diffs[pos];
      }
    }
    /* update income profile times permanent shock next period */
    double * ypnnextrow=&ypn[(age+1)*GP_SSW_NSIMS];
    ypnrow=&ypn[age*GP_SSW_NSIMS];
    _n_workrow=&_n_work[(age+1)*GP_SSW_NSIMS];
    vdMul(GP_SSW_NSIMS,ypnrow,_n_workrow,ypnnextrow);
    /* income growth is increased by ytrend */
    cblas_dscal(GP_SSW_NSIMS,params_stage1.g[age]+params_stage2.ytrend,ypnnextrow,1);
    /* update normalized cash on hand next period */
    scale=params_stage1.R/((params_stage1.g[age]+params_stage2.ytrend)*(params_stage1.f0[age]+params_stage2.ftrend));
    xsnrow=&xsn[age*GP_SSW_NSIMS];
    csnrow=&csn[age*GP_SSW_NSIMS];
    xsnnextrow=&xsn[(age+1)*GP_SSW_NSIMS];
    _u_workrow=&_u_work[(age+1)*GP_SSW_NSIMS];
    vdSub(GP_SSW_NSIMS,xsnrow,csnrow,work_N);
    vdLinearFrac(GP_SSW_NSIMS,work_N,_n_workrow,scale,0,1,0,xsnnextrow);
    vdAdd(GP_SSW_NSIMS,_u_workrow,xsnnextrow,xsnnextrow);
    rulerowstart=rulerowstart+GP_SSW_NGRID;
  }
  
  /* simulate consumption for last age */
  age=GP_SSW_T-1;
  /* interpolate consumption at each value in row age of the normalized cash on hand matrix using column age of the rule matrix */
  /* compute propensity to consume at the endpoints */
  ratioleft=rule[rulerowstart+0]*myproblempart2.xgridleftinv;
  ratioright=rule[rulerowstart+GP_SSW_NGRID-1]*myproblempart2.xgridrightinv;
  /* compute differences of the rule */
  vdSub(GP_SSW_NGRID,&rule[rulerowstart+1],&rule[rulerowstart],diffs);
  /* interpolate/extrapolate */
#pragma ivdep
  for(j=0;j<GP_SSW_NSIMS;j++) {
    double x=xsn[age*GP_SSW_NSIMS+j];
    if(x<GP_SSW_XMIN) csn[age*GP_SSW_NSIMS+j]=x*ratioleft;
    else if(x>GP_SSW_XMAX) csn[age*GP_SSW_NSIMS+j]=x*ratioright;
    else if(x<GP_SSW_XINT) {
      posd=(x-GP_SSW_XMIN)*GP_SSW_INVDELTALOW;
      pos=(int)posd;
      csn[age*GP_SSW_NSIMS+j]=rule[rulerowstart+pos]+(posd-(double)pos)*diffs[pos];
    }
    else {
      posd=(x-GP_SSW_XINT)*GP_SSW_INVDELTAHIGH;
      pos=(int)posd+GP_SSW_NGRIDLOW;
      csn[age*GP_SSW_NSIMS+j]=rule[rulerowstart+pos]+(posd-(double)(pos-GP_SSW_NGRIDLOW))*diffs[pos];
    }
  }

  /* de-normalize consumption */
  vdMul(datasize,csn,ypn,csn);
  
  /* calculate mean of log consumption by age */
#pragma ivdep
  for(age=0;age<GP_SSW_T;age++) {
    n0i=0;
    cmi=0;
    for(j=0;j<GP_SSW_NSIMS;j++) {
      if(_u_work[age*GP_SSW_NSIMS+j]!=0) {
        n0i+=1;
        cmi+=(log(csn[age*GP_SSW_NSIMS+j])-cmi)/n0i;
      }
    }
    cm[age]=exp(cmi);
  }

  /* if desired, calculate mean of log income by age */
  /* separate loop is more efficient in the usual case that we don't want mean income */
  if(calcym) {
#pragma ivdep
    for(age=0;age<GP_SSW_T;age++) {
      n0i=0;
      cmi=0;
      for(j=0;j<GP_SSW_NSIMS;j++) {
        if(_u_work[age*GP_SSW_NSIMS+j]!=0) {
          n0i+=1;
          cmi+=(log(ypn[age*GP_SSW_NSIMS+j])-cmi)/n0i;
        }
      }
      ym[age]=exp(cmi);
    }
  }

  mkl_free(_u_work);
  mkl_free(_n_work);
  mkl_free(work_NT);
  mkl_free(xsn);
  mkl_free(ypn);
  mkl_free(csn);
  mkl_free(work_N); 

}




