/* replication of Gourinchas and Parker (Econometrica 2002) */
/* sam schulhofer-wohl, 9/13/2013 */
/* computes the variance-covariance matrix for the second stage, without correction for first stage estimation */
#include <math.h>
#include <mkl.h>
#include <stdio.h>
#include "gp_ssw_intel.h"
using namespace std;

/* opw: =1 for efficient weighting matrix */
/* estctrend, estytrend, estftrend: =1 if estimated slopes in the consumption/income/famsize profiles */
/* let nparams = 4+estctrend+estytrend+estftrend. then: */
/* grad: nparams x T. must exist on input. on output, filled with gradient of each moment difference wrt beta, rho, gamma0, gamma1, ctrend (if estimated), ytrend (if estimated), ftrend (if estimated)*/
/* var2: nparams x nparams. must exist on input. on output, filled with variance-covariance matrix for beta, rho, gamma0, gamma1, ctrend (if estimated), ytrend (if estimated), ftrend (if estimated), treating first-stage parameters as fixed */
/* varall: nparams x nparams. must exist on input. on output, filled with variance-covariance matrix for beta, rho, gamma0, gamma1, ctrend (if estimated), ytrend (if estimated), ftrend (if estimated), treating first-stage parameters as estimated */
/* oid2, oidall: pointers to locations for storing the overidentification test statistic */
/* efficientweights: TxT, estimate of the efficient weighting matrix accounting for first-stage estimation */


void gp_ssw_var2(const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, const struct siminputs simshocks,
  const int opw, const int estctrend, const int estytrend, const int estftrend, 
  double *grad, double *var2, double *varall, double *efficientweights, double *oid2, double *oidall)
{
  
  int nparams=4;
  if(estctrend==1) nparams++;
  if(estytrend==1) nparams++;
  if(estftrend==1) nparams++;
  
  struct gradtol tols;
  /* don't try parameters <0 */
  tols.beta=fmin(myproblem.gradtol_beta,0.9*params_stage2.beta);
  tols.rho=fmin(myproblem.gradtol_rho,0.9*params_stage2.rho);
  tols.gamma0=fmin(myproblem.gradtol_gamma0,0.9*params_stage2.gamma[0]);
  tols.gamma1=fmin(myproblem.gradtol_gamma1,0.9*params_stage2.gamma[1]);
  tols.ytrend=myproblem.gradtol_ytrend;
  tols.ftrend=myproblem.gradtol_ftrend;
  tols.R=myproblem.gradtol_R;
  tols.p=myproblem.gradtol_p;
  tols.s_u=myproblem.gradtol_s_u;
  tols.s_n=myproblem.gradtol_s_n;

  /* produce grad here */
  int i;
  gp_ssw_dograd('b',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad[GP_SSW_T*0]);
  gp_ssw_dograd('r',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad[GP_SSW_T*1]);
  gp_ssw_dograd('g',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad[GP_SSW_T*2]);
  gp_ssw_dograd('h',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad[GP_SSW_T*3]);
  int nextparam=4;
  if(estctrend==1) {
    /* analytic gradient for ctrend */
    for(i=0;i<GP_SSW_T;i++) grad[4*GP_SSW_T+i]=-(double)i+0.5*((double)(GP_SSW_T-1));
    nextparam++;
  }
  if(estytrend==1) {
    gp_ssw_dograd('y',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad[GP_SSW_T*nextparam]);
    nextparam++;
  }
  if(estftrend==1) {
    gp_ssw_dograd('f',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad[GP_SSW_T*nextparam]);
  }

  /* workspace */
  double * varwork, * work1, * work2, * work3, * work4, * work5;
  varwork=(double *)mkl_calloc(GP_SSW_T*nparams,sizeof(double),64);
  work1=(double *)mkl_calloc(GP_SSW_T*GP_SSW_T,sizeof(double),64);
  work2=(double *)mkl_calloc(GP_SSW_T*GP_SSW_T,sizeof(double),64);
  work3=(double *)mkl_calloc(nparams*nparams,sizeof(double),64);
  work4=(double *)mkl_calloc(nparams*nparams,sizeof(double),64);
  work5=(double *)mkl_calloc(GP_SSW_T*myproblem.nparams1,sizeof(double),64);

  /* variance-covariance matrices treating first stage parameters as fixed */
  /* compute g'*w*g */
  cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasTrans,GP_SSW_T,nparams,GP_SSW_T,1,myproblem.weightmat,GP_SSW_T,grad,GP_SSW_T,0,varwork,nparams);
  cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,nparams,nparams,GP_SSW_T,1,grad,GP_SSW_T,varwork,nparams,0,var2,nparams);
  /* var2=inv(g'*w*g) */
  gp_ssw_safechol(var2,"vcv treating first stage as fixed",nparams);
  if(opw!=1) {
    /* if w was not efficient, have to do more */
    /* work3=var2[=inv(g'*w*g)]; var2 will soon be overwritten */
    cblas_dcopy(nparams*nparams,var2,1,work3,1);
    /* work1=omega*w */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,GP_SSW_T,GP_SSW_T,GP_SSW_T,1,myproblem.omegamat,GP_SSW_T,myproblem.weightmat,GP_SSW_T,0,work1,GP_SSW_T);
    /* work2=w*work1=w*omega*w */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,GP_SSW_T,GP_SSW_T,GP_SSW_T,1,myproblem.weightmat,GP_SSW_T,work1,GP_SSW_T,0,work2,GP_SSW_T);
    /* var2=g'*work2*g = g'*w*omega*w*g */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasTrans,GP_SSW_T,nparams,GP_SSW_T,1,work2,GP_SSW_T,grad,GP_SSW_T,0,varwork,nparams);
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,nparams,nparams,GP_SSW_T,1,grad,GP_SSW_T,varwork,nparams,0,var2,nparams);
    /* work4=var2*work3 = g'*w*omega*w*g * inv(g'*w*g) */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,nparams,nparams,nparams,1,var2,nparams,work3,nparams,0,work4,nparams);
    /* var2 = work3*work4 = inv(g'*w*g) * g'*w*omega*w*g * inv(g'*w*g) */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,nparams,nparams,nparams,1,work3,nparams,work4,nparams,0,var2,nparams);
  }
  /* scaling factor */
  cblas_dscal(nparams*nparams,(1+myproblem.taui)/myproblem.itm,var2,1);

  /* compute gradient wrt first-stage parameters */
  double * grad1;
  grad1=(double *)mkl_malloc(GP_SSW_T*myproblem.nparams1*sizeof(double),64);
  gp_ssw_dograd('R',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad1[GP_SSW_T*0]);
  gp_ssw_dograd('p',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad1[GP_SSW_T*1]);
  gp_ssw_dograd('u',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad1[GP_SSW_T*2]);
  gp_ssw_dograd('n',myproblem,myproblempart2,params_stage1,params_stage2,simshocks,tols,&grad1[GP_SSW_T*3]);

  /* variance-covariance matrices treating first stage parameters as estimated */
  /* compute g'*w*g */
  cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasTrans,GP_SSW_T,nparams,GP_SSW_T,1,myproblem.weightmat,GP_SSW_T,grad,GP_SSW_T,0,varwork,nparams);
  cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,nparams,nparams,GP_SSW_T,1,grad,GP_SSW_T,varwork,nparams,0,varall,nparams);
  /* varall=inv(g'*w*g) */
  gp_ssw_safechol(varall,"vcv treating first stage as estimated",nparams);
  /* efficientweights=[(1+tau)*omega+itm*g1*v1*g1']*w (useful for non-efficient vcv and oid test */
  /* )  work5=g1*v1 */
  cblas_dgemm(CblasRowMajor,CblasTrans,CblasNoTrans,GP_SSW_T,myproblem.nparams1,myproblem.nparams1,1,grad1,GP_SSW_T,myproblem.V1,myproblem.nparams1,0,work5,myproblem.nparams1);
  /* )  efficientweights=(1+tau)*omega */
  cblas_dcopy(GP_SSW_T*GP_SSW_T,myproblem.omegamat,1,efficientweights,1);
  cblas_dscal(GP_SSW_T*GP_SSW_T,1+myproblem.taui,efficientweights,1);
  /* )  efficientweights=efficientweights+itm*work5*g1'=(1+tau)*omega+itm*g1*v1*g1' */
  cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,GP_SSW_T,GP_SSW_T,myproblem.nparams1,myproblem.itm,work5,myproblem.nparams1,grad1,GP_SSW_T,1,efficientweights,GP_SSW_T);
  mkl_free(grad1);
  if(opw!=1) {
    /* if w was not efficient, have to do more */
    /* work3=varall[=inv(g'*w*g)]; varall will soon be overwritten */
    cblas_dcopy(nparams*nparams,varall,1,work3,1);
    /* work1=w*efficientweights=w*[(1+tau)*omega+itm*g1'*v1*g1] */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,GP_SSW_T,GP_SSW_T,GP_SSW_T,1,myproblem.weightmat,GP_SSW_T,efficientweights,GP_SSW_T,0,work1,GP_SSW_T);
    /* work2=work1*w=w*[(1+tau)*omega+itm*g1'*v1*g1]*w */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,GP_SSW_T,GP_SSW_T,GP_SSW_T,1,work1,GP_SSW_T,myproblem.weightmat,GP_SSW_T,0,work2,GP_SSW_T);
    /* varall=g'*work2*g = g'*w*[(1+tau)*omega+itm*g1'*v1*g1]*w*g */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasTrans,GP_SSW_T,nparams,GP_SSW_T,1,work2,GP_SSW_T,grad,GP_SSW_T,0,varwork,nparams);
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,nparams,nparams,GP_SSW_T,1,grad,GP_SSW_T,varwork,nparams,0,varall,nparams);
    /* work4=varall*work3 = g'*w*[(1+tau)*omega+itm*g1'*v1*g1]*w*g * inv(g'*w*g) */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,nparams,nparams,nparams,1,varall,nparams,work3,nparams,0,work4,nparams);
    /* varall = work3*work4 = inv(g'*w*g) * g'*w*[(1+tau)*omega+itm*g1'*v1*g1]*w*g * inv(g'*w*g) */
    cblas_dgemm(CblasRowMajor,CblasNoTrans,CblasNoTrans,nparams,nparams,nparams,1,work3,nparams,work4,nparams,0,varall,nparams);
  }
  /* scaling factor */
  cblas_dscal(nparams*nparams,1.0/myproblem.itm,varall,1);
  
  /* overidentification tests */
  gp_ssw_safechol(efficientweights,"overidentification test treating first stage as estimated",GP_SSW_T);
  *oidall=gp_ssw_dist2(myproblem,myproblempart2,params_stage1,params_stage2,simshocks,efficientweights)*myproblem.itm;
  /* efficientweights now has the efficient weighting matrix */
  cblas_dcopy(GP_SSW_T*GP_SSW_T,myproblem.omegamat,1,work1,1);
  gp_ssw_safechol(work1,"overidentification test treating first stage as fixed",GP_SSW_T);
  *oid2=gp_ssw_dist2(myproblem,myproblempart2,params_stage1,params_stage2,simshocks,work1)*myproblem.itm/(1+myproblem.taui);

  mkl_free(varwork);
  mkl_free(work1);
  mkl_free(work2);
  mkl_free(work3);
  mkl_free(work4);
  mkl_free(work5);
  
}



/* compute gradient */
/* gradout: pointer to t x 1 array, to be filled w/ gradient wrt parameter identified by par */
void gp_ssw_dograd(const char par, const struct problem myproblem, const struct problempart2 myproblempart2, const struct params1 params_stage1, const struct params2 params_stage2, const struct siminputs simshocks, const struct gradtol tols, double * gradout)
{
  struct params1 my_params_stage1_hi, my_params_stage1_lo;
  struct params2 my_params_stage2_hi, my_params_stage2_lo;
  volatile double hi,lo,partarget,tols_par;
  double scal;
  int i;
  if(par=='b') {
    partarget=params_stage2.beta;
    tols_par=tols.beta;
  }
  else if(par=='r') {
    partarget=params_stage2.rho;
    tols_par=tols.rho;
  }
  else if(par=='g') {
    partarget=params_stage2.gamma[0];
    tols_par=tols.gamma0;
  }
  else if(par=='h') {
    partarget=params_stage2.gamma[1];
    tols_par=tols.gamma1;
  }
  else if(par=='y') {
    partarget=params_stage2.ytrend;
    tols_par=tols.ytrend;
  }
  else if(par=='f') {
    partarget=params_stage2.ftrend;
    tols_par=tols.ftrend;
  }
  else if(par=='R') {
    partarget=params_stage1.R;
    tols_par=tols.R;
  }
  else if(par=='p') {
    partarget=params_stage1.p;
    tols_par=tols.p;
  }
  else if(par=='u') {
    partarget=params_stage1.s_u;
    tols_par=tols.s_u;
  }
  else if(par=='n') {
    partarget=params_stage1.s_n;
    tols_par=tols.s_n;
  }
  double * gradlowrow;
  gradlowrow=(double *)mkl_malloc(GP_SSW_T*sizeof(double),64);
  my_params_stage1_hi=params_stage1;
  my_params_stage1_lo=params_stage1;
  my_params_stage2_hi=params_stage2;
  my_params_stage2_lo=params_stage2;
  hi=partarget+tols_par/2;
  lo=partarget-tols_par/2;
  if(par=='b') {
    my_params_stage2_hi.beta=hi;
    my_params_stage2_lo.beta=lo;
  }
  else if(par=='r') {
    my_params_stage2_hi.rho=hi;
    my_params_stage2_lo.rho=lo;
  }
  else if(par=='g') {
    my_params_stage2_hi.gamma[0]=hi;
    my_params_stage2_lo.gamma[0]=lo;
  }
  else if(par=='h') {
    my_params_stage2_hi.gamma[1]=hi;
    my_params_stage2_lo.gamma[1]=lo;
  }
  else if(par=='y') {
    my_params_stage2_hi.ytrend=hi;
    my_params_stage2_lo.ytrend=lo;
  }
  else if(par=='f') {
    my_params_stage2_hi.ftrend=hi;
    my_params_stage2_lo.ftrend=lo;
  }
  else if(par=='R') {
    my_params_stage1_hi.R=hi;
    my_params_stage1_lo.R=lo;
  }
  else if(par=='p') {
    my_params_stage1_hi.p=hi;
    my_params_stage1_lo.p=lo;
  }
  else if(par=='u') {
    my_params_stage1_hi.s_u=hi;
    my_params_stage1_lo.s_u=lo;
  }
  else if(par=='n') {
    my_params_stage1_hi.s_n=hi;
    my_params_stage1_lo.s_n=lo;
  }
  gp_ssw_momentvec(myproblem,myproblempart2,my_params_stage1_hi,my_params_stage2_hi,simshocks,gradout);
  gp_ssw_momentvec(myproblem,myproblempart2,my_params_stage1_lo,my_params_stage2_lo,simshocks,gradlowrow);
  scal=1/(hi-lo);
  for(i=0;i<GP_SSW_T;i++) gradout[i]=(gradout[i]-gradlowrow[i])*scal;
  mkl_free(gradlowrow);

}


/* wrapper for cholesky inversion */
void gp_ssw_safechol(double * mat, const char * msg, const int dim)
{
  int status, safechol_i, safechol_j;
  status=LAPACKE_dpotrf(LAPACK_ROW_MAJOR,'U',dim,mat,dim);
  if(status!=0) {
    printf("cholesky decomposition failed: ");
    printf(msg);
    printf(" is invalid.\n");
  }
  else LAPACKE_dpotri(LAPACK_ROW_MAJOR,'U',dim,mat,dim);
  if(dim>1) { 
    for(safechol_i=1;safechol_i<dim;safechol_i++) for(safechol_j=0;safechol_j<safechol_i;safechol_j++) mat[safechol_i*dim+safechol_j]=mat[safechol_j*dim+safechol_i];
  } 
}

